@@ -1,6 +1,7 @@
 from rest_framework import viewsets
 
 from whisperer.models import Webhook, WebhookEvent
+from whisperer.resources.filters import WebhookEventFilter, WebhookFilter
 from whisperer.resources.serializers import WebhookEventSerializer, WebhookSerializer
 from whisperer.services import WebhookService
 
@@ -9,6 +10,7 @@ class WebhookViewSet(viewsets.ModelViewSet):
     queryset = Webhook.objects.all()
     serializer_class = WebhookSerializer
     service = WebhookService()
+    filter_class = WebhookFilter
 
     def get_queryset(self):
         queryset = super(WebhookViewSet, self).get_queryset()
@@ -32,6 +34,7 @@ class WebhookViewSet(viewsets.ModelViewSet):
 
 class WebhookEventViewSet(viewsets.ReadOnlyModelViewSet):
     queryset = WebhookEvent.objects.all()
+    filter_class = WebhookEventFilter
     serializer_class = WebhookEventSerializer
 
     def get_queryset(self):
